'''
Created on Jul 10, 2025

@author: ahypki
'''

class Logger:
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    RED = '\033[91m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    
    DEBUG = False
    INFO = True
    
    def logDebug(msg, printLogLevel = True, printNewLine = True):  # @NoSelf
        if Logger.DEBUG:
            print(("DEBUG " if printLogLevel else '') 
                  + str(msg), 
                  end = ('\n' if printNewLine else ''))
    
    def logInfo(msg, printLogLevel = True, printNewLine = True):  # @NoSelf
        if Logger.INFO:
            print(("INFO  " if printLogLevel else '') 
                  + str(msg), 
                  end = ('\n' if printNewLine else ''))
        
    def logWarn(msg, printLogLevel = True, printNewLine = True):  # @NoSelf
        print(("WARN  " if printLogLevel else '') 
              + str(msg), 
              end = ('\n' if printNewLine else ''))
    
    def logError(msg, printLogLevel = True, printNewLine = True):  # @NoSelf
        print(("ERROR " if printLogLevel else '') 
              + str(msg), 
              end = ('\n' if printNewLine else ''))



'''
Created on Jul 15, 2025

@author: ahypki
'''

class Settings(object):
    '''
    classdocs
    '''

    MAX_COMMIT_BYTES = 5_000_000
    
    SEPARATOR_NEXT_REPO = '############################################################'
    PREFIX_LINE = "# "

    def __init__(self, params):
        '''
        Constructor
        '''
        


'''
Created on Jul 15, 2025

@author: ahypki
'''

import sys

def getArgString(name, defaultvalue):
    for i in range(len(sys.argv)):
        if sys.argv[i] == "--" + name:
            return sys.argv[i + 1]
    return defaultvalue

def isArgPresent(name):
    for i in range(len(sys.argv)):
        if sys.argv[i] == "--" + name:
            return True
    return False


'''
Created on Jul 15, 2025

@author: ahypki
'''
from gima.utils.Logger import Logger
import subprocess
from gima.utils.regex import firstGroup

def linux_cmd(cmd, workingDir = None, printImmediatelly = False):
    if printImmediatelly:
        Logger.logDebug("linux cmd: " + cmd);
    result = []
    s = subprocess.Popen(cmd, cwd = workingDir, shell=True, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    for line in s.stdout.readlines():
        strLine = str(line).rstrip()
#         if printImmediatelly is False:
        result.append(firstGroup(strLine, 'b\'(.*)\\\\n\''))
#         else:
        if printImmediatelly:
            print(firstGroup(strLine, 'b\'(.*)\\\\n\''))
    return result


'''
Created on Jul 10, 2025

@author: ahypki
'''
import json
from pathlib._local import PosixPath

from gima.utils.Logger import Logger


class Database:
    __database_path = PosixPath('~/.gim.json').expanduser()
    
    __initiated = False
    __data = None
    
    def __init__(self):
        if (self.__initiated):
            return
        
        self.__initiated = True
        Logger.logDebug('Reading gim database {}'.format(self.__database_path))
        
        with open(self.__database_path, 'r') as file:
            self.__data = json.load(file)
            
    def __getRepo(self, path):
        for gitRepo in self.__data['gitRepos']:
            if gitRepo['path'] == path:
                return gitRepo
        return None
            
    def iterateRepos(self, group):        
        repos = [];
        for gitRepo in self.__data['gitRepos']:
            if group is not None:
                # checking if the git repo in in the given group
                if group in gitRepo['groups']:
                    repos.append(gitRepo['path'])
            else:
                repos.append(gitRepo['path'])
        return repos
    
    def printSummary(self):       
        Logger.logDebug('gim summary:')
        
        # repositories
        for gitRepo in self.__data['gitRepos']:
            groups = ''
            for g in gitRepo['groups']:
                groups += (', ' if len(groups) > 0 else '') + str(g)
            Logger.logDebug("\t\tRepository: " + gitRepo['path'] + " (" + groups + ")")
            
        # ignored paths
        for ign in self.__data['ignorePaths']:
            Logger.logDebug("\t\tIgnored: " + ign)
        for ign in self.__data['ignoreSubPaths']:
            Logger.logDebug("\t\tIgnored: " + ign)
            
    def isRepoPresent(self, testPath):
        for gitRepo in self.__data['gitRepos']:
            if gitRepo['path'] == testPath:
                return True
        return False
    
    def isIgnored(self, testPath):
        for ign in self.__data['ignorePaths']:
            if ign == testPath:
                return True
        for ign in self.__data['ignoreSubPaths']:
            if ign == testPath or testPath.startswith(ign):
                return True
        return False

    def save(self):
        with open(self.__database_path, 'w', encoding='utf-8') as f:
            json.dump(self.__data, f, ensure_ascii=False, indent=4)
        Logger.logDebug('Saved gim database {}'.format(self.__database_path))
        
    def addRepo(self, repo):
        if repo not in self.__data['gitRepos']:
            repoJson = {}
            repoJson['path'] = repo
            repoJson['groups'] = []
            repoJson['groups'].append('1')
            self.__data['gitRepos'].append(repoJson)
            Logger.logInfo("{} added to database".format(Logger.BLUE + repo + Logger.ENDC))
            self.save()
    
    def ignore(self, ignoreDir):
        if ignoreDir not in self.__data['ignorePaths']:
            self.__data['ignorePaths'].append(ignoreDir)
            Logger.logInfo("{} added to ignored folders".format(ignoreDir))
            self.save()
            
    def ignoreAll(self, ignoreDir):
        if ignoreDir not in self.__data['ignoreSubPaths']:
            self.__data['ignoreSubPaths'].append(ignoreDir)
            Logger.logInfo("{} added to ignored folders (and all its subfolders)".format(ignoreDir))
            self.save()


'''
Created on Jul 15, 2025

@author: ahypki
'''

import os.path
import re
import subprocess
import sys
import time
# import urllib.request

def firstGroup(s, regex):
    r = re.compile(regex)
    m = r.search(s)
    if m:
        return m.group(1)
    return ""

def isNumber(s):
    return matches(s, '^[\d]+$')

def matches(s, regex):
    r = re.compile(regex)
    m = r.search(s)
    if m:
        return True
    return False


'''
Created on Jul 10, 2025

@author: ahypki
'''
import os
from sys import path
import sys

from gima.utils.args import getArgString, isArgPresent
from gima.git.Settings import Settings
from gima.utils.Logger import Logger
from gima.db.Database import Database
# from gima.git.GitRepo import GitRepo
from gima.git.GitRepo import GitRepo


database = Database()

def prompt(gitRepo, readCommand = False):
    path = gitRepo.getPath()
    status = None
    if gitRepo.isClean():
        status = Logger.GREEN + 'clean' + Logger.ENDC
    else:
        status = Logger.WARNING + gitRepo.statusShort() + Logger.ENDC 
    
    prompt = Settings.PREFIX_LINE\
            + Logger.BLUE + 'gim > ' + Logger.ENDC \
            + ('[' + (status).rjust(18, " ") + '] ' if status is not None else '') \
            + Logger.BLUE + str(path) + Logger.ENDC \
            + ' > '
    if readCommand:
        print(prompt, end = '', flush = True)
        return sys.stdin.readline().strip()
    else:
        Logger.logInfo(prompt, printLogLevel = False)

def printUsage():
    Logger.logInfo(Logger.BLUE + 'gim usage:' + Logger.ENDC, printLogLevel=False)
    Logger.logInfo('\t{} - prints summary'.format((Logger.GREEN + '--summary' + Logger.ENDC).rjust(18, ' ')), printLogLevel=False)
    Logger.logInfo('\t{} - interactively make a commit'.format((Logger.GREEN + '--commit [--group ... --path ...]' + Logger.ENDC).rjust(18, ' ')), printLogLevel=False)
    Logger.logInfo('\t\t{} - add file(s) by id, idFrom-idTo or using wildcard pattern'.format((Logger.GREEN + 'a pattern' + Logger.ENDC).ljust(18, ' ')), printLogLevel=False)
    Logger.logInfo('\t\t{} - ignore file(s) by id, idFrom-idTo or using wildcard pattern'.format((Logger.GREEN + 'i pattern' + Logger.ENDC).ljust(18, ' ')), printLogLevel=False)
    Logger.logInfo('\t\t{} - commit'.format((Logger.GREEN + 'c' + Logger.ENDC).ljust(18, ' ')), printLogLevel=False)
    Logger.logInfo('\t\t{} - commit and then push'.format((Logger.GREEN + 'cp' + Logger.ENDC).ljust(18, ' ')), printLogLevel=False)
    Logger.logInfo('\t\t{} - push only'.format((Logger.GREEN + 'push' + Logger.ENDC).ljust(18, ' ')), printLogLevel=False)
    Logger.logInfo('\t\t{} - pull only'.format((Logger.GREEN + 'pull' + Logger.ENDC).ljust(18, ' ')), printLogLevel=False)
    Logger.logInfo('\t\t{} - go to the next repository'.format((Logger.GREEN + 'n' + Logger.ENDC).ljust(18, ' ')), printLogLevel=False)
    Logger.logInfo('\t\t{} - quit'.format((Logger.GREEN + 'q' + Logger.ENDC).ljust(18, ' ')), printLogLevel=False)
    Logger.logInfo('\t{} - scans for git repos in the current folder or in the folder specified with --param'.format((Logger.GREEN + '--scan [--path ...]' + Logger.ENDC).rjust(18, ' ')), printLogLevel=False)
    
def commitOneRepo(repo):
    while True:
        Logger.logInfo(Settings.SEPARATOR_NEXT_REPO, printLogLevel= False)
        gitRepo = GitRepo(repo)
        if gitRepo.isGitRepo():
            items, size = gitRepo.status()
            if (len(items) > 0):
                cmd = prompt(gitRepo, True)
                
                if cmd.startswith('cp '):
                    msg = cmd[2:].strip()
                    gitRepo.commit(msg)
                    gitRepo.push()
                elif cmd.startswith('c '):
                    msg = cmd[2:].strip()
                    gitRepo.commit(msg)
                elif cmd.startswith('a '):
                    if size > Settings.MAX_COMMIT_BYTES:
                        Logger.logWarn('Already added {} MB in this commit, skipping'.format(str(size / 1_000_000)))
                    else:
                        pattern = cmd[1:].strip()
                        gitRepo.add(pattern)
                elif cmd.startswith('i '):
                    pattern = cmd[1:].strip()
                    gitRepo.ignore(pattern)
                elif cmd.startswith('push'):
                    gitRepo.push()
                elif cmd.startswith('pull'):
                    gitRepo.pull()
                elif cmd.startswith('l'):
                    pass
                elif cmd.startswith('n'):
                    break
                elif cmd.startswith('q'):
                    # break checking other repos
                    return True
                else:
                    Logger.logError('Unknown command: {}'.format(cmd))
            else:
                prompt(gitRepo)
                break
        else:
            # not a git repo
            Logger.logError("Not a git repo in {}".format(repo))
            break
        
def commit():
    if isArgPresent("path"):
        commitOneRepo(getArgString("path", None))
    else:
        for repo in database.iterateRepos(getArgString('group', None)):
            if (os.path.exists(repo)):
                if commitOneRepo(repo):
                    break

def scan():
    currentDir = os.getcwd()
    if getArgString('path', None) is not None:
        currentDir = getArgString('path', None)
        
    foldersSkip = []
    foldersCountAll = 0
    foldersCountIgnored = 0
    foldersCountGitRepos = 0
    
    Logger.logInfo('Scanning {}...'.format(currentDir))
        
    for subdir, dirs, files in os.walk(currentDir):
        subdir = str(subdir)
        
        # statistics
        foldersCountAll += 1
        if (foldersCountAll == 100
            or foldersCountAll == 1000 
            or foldersCountAll % 10000 == 0):
            Logger.logInfo("{} folders scanned, {} ignored, {} already tracking".format(foldersCountAll, foldersCountIgnored, foldersCountGitRepos))
        
        # check if folder was checked already
        skip = False
        for ign in foldersSkip:
            if subdir.startswith(ign):
                # logDebug('Already checked ' + subdir)
                skip = True
                break
        if skip:
            continue
                    
        # check if the folder is ignored in the database
        if database.isIgnored(subdir):
            gitRepo = GitRepo(subdir)
            if gitRepo.isGitRepo():
                foldersSkip.append(subdir)
                foldersCountIgnored += 1
            # logDebug('Ignored in database ' + subdir)
            continue
            
        gitRepo = GitRepo(subdir)
        if gitRepo.isGitRepo():
            subdir = str(gitRepo.getPath())
            if subdir not in foldersSkip and not database.isRepoPresent(subdir):
            #     logInfo("Git repo in {} is ignored".format(subdir))
            # else:
                Logger.logInfo("{} > [A]dd/[I]gnore/ignore [R]ecursivelly/[S]kip for now".format(Logger.BLUE + subdir + Logger.ENDC))
                cmd = sys.stdin.readline().strip()
                if cmd.lower() == "a":
                    database.addRepo(subdir)
                elif cmd.lower() == "i":
                    database.ignore(subdir)
                elif cmd.lower() == "r":
                    database.ignoreAll(subdir)
                elif cmd.lower() == "s":
                    pass
            else:
                foldersCountGitRepos += 1

            foldersSkip.append(str(subdir))
    Logger.logInfo("{} folders scanned, {} ignored, {} already tracking".format(foldersCountAll, foldersCountIgnored, foldersCountGitRepos))
    
def main():
    if isArgPresent('verbose'):
        DEBUG = True
    
    if (isArgPresent('summary')):
        database.printSummary(None)
    elif (isArgPresent('ignore')):
        database.ignore(getArgString('ignore', None))
    elif (isArgPresent('ignore-all')):
        database.ignoreAll(getArgString('ignore-all', None))
    elif (isArgPresent('commit')):
        commit()
    elif (isArgPresent('scan')):
        scan()
    else:
        printUsage()
        
    Logger.logInfo('Finished!')
    
if __name__ == '__main__':
    main()


from setuptools import setup

setup(
    name='gima',
    version='1.0.7',
    author='Arkadiusz Hypki',
    description='Software which simplifies managing many git repositories through a console',
    packages=['gima', 'gima/db', 'gima/git', 'gima/utils'],
    include_package_data=True,
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        'pygit2'
    ],
    entry_points = {
        'console_scripts': [
            'gima = gima.gima:main'
        ]
    },
)



'''
Created on 15-06-2013

@author: ahypki
'''
from fileinput import filename
import fnmatch
import os
from pathlib._local import Path, PurePosixPath

import pygit2
from pygit2.enums import FileStatus
from pygit2.repository import Repository

from gima.git.Settings import Settings
from gima.utils.regex import matches, isNumber
from gima.utils.Logger import Logger
from gima.utils.io import linux_cmd


class GitRepo:
    __path = None
    
    def __init__(self, path):
        self.__path = path
        
    def add(self, pattern):
        if matches(pattern.strip(), '[0-9\\-\\s]+'):
            self.addByIndices(pattern)
        else:
            self.addByWildcards(pattern)
    
    def ignore(self, pattern):
        if matches(pattern.strip(), '[0-9\\-\\s]+'):
            self.ignoreByIndices(pattern)
        else:
            self.ignoreByWildcards(pattern)
            
    def addByIndices(self, pattern):
        if (isNumber(pattern)):
            toAdd = int(pattern)
            i = 1
            items, size = self.status()
            for filename, code in items:
                if (i == toAdd):
                    self.addFile(filename)
                    break
                i += 1
        elif '-' in pattern:
            fromTo = pattern.split('-')
            fromIdx = int(fromTo[0])
            toIdx = int(fromTo[1])
            addedBytes = 0
            i = 1
            items, size = self.status()
            for filename, code in items:
                if (i >= fromIdx and i <= toIdx):
                    self.addFile(filename)
                    addedBytes += os.path.getsize(self.__path + '/' + filename)
                i += 1
                if addedBytes > Settings.MAX_COMMIT_BYTES:
                    Logger.logWarn("Already added {} MB, breaking adding more".format(str(addedBytes / 1_000_000)))
                    break
                
    def ignoreByIndices(self, pattern):
        if (isNumber(pattern)):
            toAdd = int(pattern)
            i = 1
            items, size = self.status()
            for filename, code in items:
                if (i == toAdd):
                    self.ignoreFile(filename)
                    break
                i += 1
        elif '-' in pattern:
            fromTo = pattern.split('-')
            fromIdx = int(fromTo[0])
            toIdx = int(fromTo[1])
            i = 1
            items, size = self.status()
            for filename, code in items:
                if (i >= fromIdx and i <= toIdx):
                    self.ignoreFile(filename)
                i += 1
    
    def addByWildcards(self, pattern):
        items, size = self.status()
        for filename, code in items:
            if (code == FileStatus.INDEX_NEW 
                or code == FileStatus.WT_NEW
                or code == FileStatus.WT_DELETED
                or code == FileStatus.WT_MODIFIED):
                if fnmatch.fnmatch(filename, pattern):
                    self.addFile(filename)
    
    def ignoreByWildcards(self, pattern):
        items, size = self.status()
        for filename, code in items:
            if (code == FileStatus.INDEX_NEW 
                or code == FileStatus.WT_NEW
                or code == FileStatus.WT_DELETED
                or code == FileStatus.WT_MODIFIED):
                if fnmatch.fnmatch(filename, pattern):
                    self.ignoreFile(filename)
        
    def toString(self, code):
        if code == FileStatus.WT_MODIFIED:
            return "MODIFIED"
        elif code == FileStatus.WT_NEW:
            return "??"
        elif code == FileStatus.INDEX_NEW:
            return "NEW"
        elif code == FileStatus.WT_DELETED:
            return "DELETED"
        elif code == FileStatus.INDEX_DELETED:
            return "DELETED"
        else:
            return str(code)
    
    def addFile(self, filename):
        repo = Repository(self.__path)
        repo.index.add(filename)
        repo.index.write()
        
    def ignoreFile(self, filename):
        with open(self.__path + '/.gitignore', 'a') as content_file:
            content_file.write('\n')
            content_file.write(filename)
    
    def commit(self, msg):
        repo = Repository(self.__path)
        
        # adding all modified files
        # repo.index.add_all() - SUPER slow!
        items, size = self.status()
        for filename, code in items:
            if (code == FileStatus.INDEX_NEW 
                # or code == FileStatus.WT_NEW
                or code == FileStatus.WT_DELETED
                or code == FileStatus.WT_MODIFIED):
                    if os.path.exists(self.__path + '/' + filename):
                        repo.index.add(filename)
                    else:
                        repo.index.remove(filename)
        
        repo.index.write()
        tree = repo.index.write_tree()
        author = pygit2.Signature(repo.config.__getitem__('user.name'), 
                                  repo.config.__getitem__('user.email'))
        committer = pygit2.Signature(repo.config.__getitem__('user.name'), 
                                     repo.config.__getitem__('user.email'))
        ref = repo.head.name
        parents = [ repo.head.target ]
        repo.create_commit(
            ref,
            author,
            committer,
            msg,
            tree,
            parents,
        )
        
    def isGitRepo(self):
        try:
            repo = Repository(self.__path)
            items = repo.status().items()
            return True
        except Exception as e:
            return False
        
    def status(self, printToScreen = True):
        repo = Repository(self.__path)
        items = repo.status().items()
        i = 1
        size = 0
        for filename, code in items:
            if printToScreen:
                print(Settings.PREFIX_LINE + ('[' + str(i) + ']').rjust(5, " ")
                      + ' ' 
                      + self.toString(code).rjust(10, " ")
                      + ' ' 
                      + filename
                      + ' [' + str(i) + ']')
                i += 1
            if (code == FileStatus.INDEX_NEW 
                # or code == FileStatus.WT_NEW
                # or code == FileStatus.WT_DELETED
                or code == FileStatus.WT_MODIFIED):
                try:
                    size += os.path.getsize(self.__path + '/' + filename)
                except Exception as e:
                    pass
        return items, size
    
    def push(self):
        try:
            # logger.info("Pushing changes to git")
            repo = Repository(self.__path)
            remote = "origin"
            branch = "refs/heads/" + repo.head.shorthand
            sshcred = pygit2.Keypair('git',
                                       Path('~/.ssh/id_rsa.pub').expanduser(),
                                       Path('~/.ssh/id_rsa').expanduser(),
                                       '')
            repo.crendentals = sshcred
            # credentials = pygit2.UserPass(userName, password)
            callbacks = pygit2.RemoteCallbacks(credentials=sshcred)
            repo.remotes[remote].push(['+' + branch], callbacks=callbacks)
            
            Logger.logInfo('gim > ' + Logger.GREEN + str(self.__path) + Logger.ENDC + ' > '
                + ' [' + Logger.GREEN + 'Push OK' + Logger.ENDC + ']')
        except Exception as e:
            try:
                linux_cmd('git push', self.__path, False)
            except Exception as e:
                Logger.logError(e)
            
    def isClean(self):
        repo = Repository(self.__path)
        items = repo.status().items()
        if len(items) == 0:
            return True
        else:
            return False
    
    def statusShort(self):
        if self.isClean():
            return "clean"
        else:
            try:
                repo = Repository(self.__path)
                upstream_head = repo.revparse_single('origin/HEAD')
                local_head    = repo.revparse_single('HEAD')
            
                diff = repo.ahead_behind(local_head.id, upstream_head.id)
                if diff[0] > 0:
                    return "ahead " + str(diff[0])
                elif diff[1] > 0:
                    return "behind " + str(diff[1])
                else:
                    status = "modified"
                    items, size = self.status(printToScreen = False)
                    sizeMB = size / 1_000_000
                    return status + ", " + ("{:.1f}".format(sizeMB)) + " MB"
            except Exception as e:
                return "modified"
            
    def getPath(self):
        repo = Repository(self.__path)
        return PurePosixPath(repo.path).parent
    
    def pull(self):
        try:
            linux_cmd('git pull', self.__path, False)
        except Exception as e:
            Logger.logError(e)
    
    
    
